/*****************************************************************************

 JOCO's ZX spectrum emulator.
 sp1.c: the spectrum emulator main controller

 see the file read.me for details/legal

******************************************************************************/
#include <stdio.h>
#include <dpmi.h>
#include <time.h>

#include "zx_hw.h"
#include "z80cpu.h"

/* globals */
static dword addr;
static byte  data;
 
/* function implemented in the tapes.c file */
extern void CloseTape();
extern int OpenTape(const char *name);
extern int AllocTapeBuff();
extern void FreeTapeBuff();

/* write a special byte. Called if the address is in the range
   0x0000..0x5B00, this is the rom/screen/attr area of the spectrum */
/* %edi: address, %al: byte to write to */
void SpecWriteByte() {
 __asm("
	pushal
	movl	%edi,_addr
	movb	%al,_data
 ");
 /* if it's the screenram */
 if (addr>16383) {
   Z80mem[addr]=data; /* update the real spectrum memory, too */
   /* screenbyte(Sput) or attributebyte(Aput) update */
   if (addr<22528+768) (addr<22528)?(Sput(addr,data)):(Aput(addr,data));
 }
 __asm("popal");
}

/* write spec.word: it calls the SpecWriteByte twice for the l/h bytes */
/* %edi: address, %ax: word to write to */
void SpecWriteWord() {
 __asm("
	pushl	%eax
	call	_SpecWriteByte
	movb	%ah,%al
	inc	%edi
	call	_SpecWriteByte
	dec	%edi
	popl	%eax	
");
}

/* Spectrum IN ports realization (ula: bit #0 is 0) */
void ByteIn() {
 __asm("
	mov	$255,%al
	test	$1,%cl
	jnz	in_quit
	call	_ULAin
in_quit:
 ");
}

/* Spectrum OUT ports realization (ula: bit #0 is 0) */
void ByteOut() {
 __asm("
	test	$1,%cl
	jnz	out_quit
	call	_ULAout
out_quit:
 ");
}

void ExtSub() {
}

/* ZX ROM patch codes: tape handling */

void ExtProc0() {
 __asm("
 call _LoadFile
 ");
}

void ExtProc1() {
}

void ExtProc2() {
}

void ExtProc3() {
}

void delay(int d) {
 time_t	a;
 a=time(NULL);
 while (a+d<time(NULL));
}

void Abort() {
   UninstallKeyboardISR();
   UninstallTimerISR();
   PcScreen();
}

/*************************************************************************/
/*  Here we go. The emulator starts 					 */
/*************************************************************************/
int main() {
 FILE *f;
 int   i;

 sound(440);

 /* Create a descriptor for the PC screen area */
 DescA000=__dpmi_segment_to_descriptor(0xa000);
 /* Initalize the spectrum screenmode */
 SpectrumScreen();
 /* Install the timer ISR (50IRQS/sec) */
 InstallTimerISR(); 
 /* Clear the keyboard table */
 memset(PCKBD,0x00,256);
 /* Install the keyboard ISR */
 InstallKeyboardISR();

 /* memory allocation:
    64K for the normal spectrum RAM
    64K for the memory specification (RAM/ROM/Spec RAM)
    4 bytes for "padding" ( just in case Z80 PC=0xffff )
 */
 Z80mem=(byte *)malloc(2*(65536+4));
 /* memory specifier */
 Z80memSpec=Z80mem+65536+4;
 /* clear both */
 memset(Z80mem,0x00,65536+4);
 memset(Z80memSpec,0x00,65536+4);

 /* special memory array: ROM(16K),SCREEN(6K),ATTRIBUTES(768b) */
 /* because of the specification, the CPU emulator will call the
    SpecWriteByte and SpecWriteWord functions when Z80 wants to
    write into this memory loactions */
 memset(Z80memSpec,0x01,16384+6144+768);

 /* read the ROM image */
 if (f=fopen("zxrom.bin","rb")) {
   fread(Z80mem,1,16384,f);
   fclose(f);
 } else {
   Abort();
   printf("The default ROM image 'zxrom.bin' not found. Abort.\n");
   return 1;
 }

 // tape patches
 Z80mem[0x0556]=0x21; // ld hl,$053f 
 Z80mem[0x0557]=0x3f;
 Z80mem[0x0558]=0x05;
 Z80mem[0x0559]=0xE5; // push hl     
 Z80mem[0x055A]=0xED; // extproc $00  
 Z80mem[0x055B]=0x00;
 Z80mem[0x055C]=0xC9; // ret          

 /* Allocate the buffer, then open the tape file */
 AllocTapeBuff();
 if (!OpenTape("testtape.tap")) {
  FreeTapeBuff();
  Abort();
  printf("Default tapefile 'testtape.tap' not found. Abort.\n");
  return 2;
 } 

 /* Reset the CPU */
 Z80Reset();

 /* Main execution cycle */
 do {
  /* Execute one Z80 instruction */
  ExecOne();
  /* some delay: calibrated to a 2.4GHz intel proc (more or less) */
  __asm("
	pushl	%ecx
	movl	$3000,%ecx
ml0:
	nop
	nop
	loop	ml0
	
	popl	%ecx
  ");
  /* F10: HARD reset spectrum */
  if (PCKBD[68]) Z80Reset();

 } while (!PCKBD[1]); /* ESC:  quit */

 /* close the tapefile and free the buffer */
 CloseTape();
 FreeTapeBuff();

 /* Deinstall ISRs, restore PC screen (80x25/text) */
 Abort();

 return 0;
}